import { NS, Server } from "@ns"
import { info } from "console";
import { calculateServerGrowth, FormulaHelper } from "./formula";
import { Scheduler } from "./scheduler";

export type BatchType = "hack" | "grow" | "weaken" | "weakenLoop";

type BatchFilePackEntry = {
    type: BatchType;
    path: string;
    content: string;
    ram: number;
};

type FilePack = BatchFilePackEntry[];

const createScriptContent: (f: string) => string = (funcName: string) => `export async function main(ns) {
    if(ns.args.length < 3){
        ns.print("[ERROR] args required");
        ns.print("run cmd [sleepDuration] [hostname] [effectStock]")
        ns.exit();
        return;
    }
    const sleepDuration = parseInt(ns.args[0]);
    const hostname = ns.args[1].toString();
    const stock = ns.args[2] == "true";
    if(sleepDuration > 0){
        await ns.sleep(sleepDuration);
    }
    await ns.${funcName}(hostname,{
        stock,
    });
}`

async function unpackFileEntry(ns: NS, filePack: BatchFilePackEntry): Promise<void> {
    await ns.write(filePack.path, filePack.content, "w");
}

export async function unpackFilePack(ns: NS, pack: FilePack): Promise<void> {
    for (const entry of pack) {
        await unpackFileEntry(ns, entry);
    }
}

export const batchFilePack: FilePack = [
    {
        type: "hack",
        path: "/tmp/hack.js",
        content: createScriptContent("hack"),
        ram:1.7,
    },
    {
        type: "weaken",
        path: "/tmp/weaken.js",
        content: createScriptContent("weaken"),
        ram: 1.75,
    },
    {
        type: "grow",
        path: "/tmp/grow.js",
        content: createScriptContent("grow"),
        ram: 1.75,
    },
    {
        type:"weakenLoop",
        path: "/tmp/weakenLoop.js",
        ram: 1.75,
        content: `export async function main(ns) {
    if(ns.args.length < 3){
        ns.print("[ERROR] args required");
        ns.print("run cmd [sleepDuration] [hostname] [effectStock]")
        ns.exit();
        return;
    }
    const sleepDuration = parseInt(ns.args[0]);
    const hostname = ns.args[1].toString();
    const stock = ns.args[2] == "true";
    for(;;){
        await ns.weaken(hostname,{
            stock,
        });
    }
}`,
    }
]

export async function scpBatchFilePack(ns: NS, hostname: string): Promise<void> {
    const list = batchFilePack.map(entry => entry.path);
    await installBatchFilePack(ns);
    await ns.scp(list, hostname);
}

export async function installBatchFilePack(ns: NS): Promise<void> {
    await unpackFilePack(ns, batchFilePack);
}

export function isBatchFilePackInstalled(ns: NS, hostname?: string): boolean {
    return batchFilePack.every(entry => ns.fileExists(entry.path, hostname))
}

export interface ExecOption {
    hostname: string;
    /**
     * @default 1
     */
    numThread?: number;
    /**
     * milliseconds to sleep before exec
     * @default 0
     */
    sleepDuration?: number;
    /**
     * target hostname to operate "hack", "weaken" or "grow"
     */
    target: string;
    stock?: boolean;
}

function execBatchfile(ns: NS, path: string, option: ExecOption): number {
    const uuid = crypto.randomUUID();
    return ns.exec(path, option.hostname, option.numThread,
        option.sleepDuration ?? 0, option.target, option.stock ?? false, uuid);
}

export function execHack(ns: NS, option: ExecOption): number {
    return execBatchfile(ns, "/tmp/hack.js", option);
}
export function execGrow(ns: NS, option: ExecOption): number {
    return execBatchfile(ns, "/tmp/grow.js", option);
}
export function execWeaken(ns: NS, option: ExecOption): number {
    return execBatchfile(ns, "/tmp/weaken.js", option);
}
export function execWeakenLoop(ns: NS, option: ExecOption): number {
    return execBatchfile(ns, "/tmp/weakenLoop.js", option);
}
export function execScriptOf(ns: NS,type: BatchType, option:ExecOption):number {
    return execBatchfile(ns, "/tmp/" + type + ".js", option);
}
export function getRamOfScript(ns: NS, type: BatchType): number{
    return batchFilePack.filter(x=>x.type == type)[0].ram;
}

type ReserveRequest = {
    hostname: string;
    usableRam: number;
}

function getKeyOfHWGWBatchManager(hostname: string): string{
    return `HWGW-${hostname}`;
}
interface HWGW {
    HThreadCount: number;
    HWThreadCount: number;
    GThreadCount: number;
    GWThreadCount: number;
}

function getTotalRamOf(hwgw: HWGW): number{
    return hwgw.HThreadCount * 1.7 + hwgw.HWThreadCount * 1.75 + hwgw.GThreadCount * 1.75 + hwgw.GWThreadCount * 1.75;
}

type Resource = {
    hostname: string;
    usableRam: number;
};

interface HackCalculation extends HWGW {
    earnMoneyPerCycle: number;
    earnExpPerCycle: number;
    /**
     * cycle time in milliseconds.
     */
    cycleTime: number;
    totalThreadCount: number;

    weakenTime: number;
    growTime: number;
    hackTime: number;
}

export function calculateBatchResource(ns: NS, hostname: string, hackThread: number, startMoney?: number, cores = 1): HackCalculation| null {
    const helper = new FormulaHelper(ns);
    const server = ns.getServer(hostname);
    startMoney ??= server.moneyMax;
    server.hackDifficulty = server.minDifficulty;

    const percentMoneyHacked = helper.calculatePercentMoneyHacked(server);
    const hackingChance = helper.calculateHackingChance(server);
    const earnMoneyPerCycle = hackThread * percentMoneyHacked * startMoney * hackingChance;
    const earnExpPerCycle = helper.calculateHackingExpGain(server) * hackThread * hackingChance;
    const cycleTime = 1000 + helper.calculateWeakenTime(server);
    const increasedSecurityAfterHack = helper.calculateHackSecurity(hackThread) + server.minDifficulty;
    const nWeakenHackThreads = Math.ceil(helper.calculateWeakenThreadCountToTargetSecurity(
        server.minDifficulty,
        increasedSecurityAfterHack, cores));
    if(startMoney <= earnMoneyPerCycle){
        return null;
    }
    const nGrowThreads = Math.ceil(helper.calculateGrowThreadCount(hostname, startMoney - earnMoneyPerCycle, startMoney, cores));
    const increasedSecurityAfterGrow = helper.calculuateGrowthSecurity(nGrowThreads, cores) + server.minDifficulty;

    const nWeakenGrowThreads = Math.ceil(helper.calculateWeakenThreadCountToTargetSecurity(
        server.minDifficulty,
        increasedSecurityAfterGrow, cores));

    return {
        weakenTime: helper.calculateWeakenTime(server),
        growTime: helper.calculateGrowTime(server),
        hackTime: helper.calculateHackingTime(server),
        earnMoneyPerCycle,
        earnExpPerCycle,
        cycleTime,
        totalThreadCount: hackThread + nWeakenHackThreads + nGrowThreads + nWeakenGrowThreads,
        HThreadCount: hackThread,
        HWThreadCount: nWeakenHackThreads,
        GThreadCount: nGrowThreads,
        GWThreadCount: nWeakenGrowThreads
    }
}
class BatchManager {
    server: Server;
    ns: NS;
    totalRam: number
    scheluder: Scheduler;
    resources: Resource[] = [];

    constructor(ns: NS, hostname: string) {
        this.ns = ns;
        this.server = ns.getServer(hostname);
        this.totalRam = 0;
        this.scheluder = new Scheduler(ns, 250);
    }

    reserve(hostname: string, usableRam: number): void {
        this.resources.push({
            hostname,
            usableRam,
        });
        this.totalRam += usableRam;
    }

    getUsableTotalRam(): number {
        return this.resources.reduce((acc, cur) => acc + cur.usableRam, 0);
    }

    execute(type: BatchType, opt: Omit<ExecOption, "hostname">): number[] {
        const ret = [];
        let numThread = opt.numThread ?? 1;
        const ramUnit = getRamOfScript(this.ns, type);
        for (let i = 0; i < this.resources.length; i++) {
            const resource = this.resources[i]
            let threadToRun = Math.floor(resource.usableRam / ramUnit);
            if (threadToRun > numThread) {
                threadToRun = numThread;
            }
            if (threadToRun <= 0) continue;

            const pid = execScriptOf(this.ns, type, {
                hostname: resource.hostname,
                numThread: threadToRun,
                sleepDuration: opt.sleepDuration,
                target: opt.target,
                stock: opt.stock,
            });
            resource.usableRam -= threadToRun * ramUnit;
            // eslint-disable-next-line @typescript-eslint/explicit-function-return-type
            const doWork = ()=>{
                resource.usableRam += threadToRun * ramUnit;
                const msg = [
                    `${pid} has finished.`,
                    `${resource.hostname} recover ${threadToRun * ramUnit} ram`]
                this.ns.print(msg.join("\n"));
            }
            this.scheluder.onPidTerminate(pid, {work() {doWork();}});
            numThread -= threadToRun;
            ret.push(pid);
        }

        return ret;
    }
}

class HWGWBatchManager extends BatchManager {
    constructor(ns: NS, hostname: string) {
        super(ns, hostname);
    }

    calcBatchResource(hackThread: number): HackCalculation| null {
        return calculateBatchResource(this.ns, this.server.hostname, hackThread, this.server.moneyMax)
    }
    /**
     * execute batch
     * do not await this function.
     * @param hwgw HackCalculation of HWGW
     * @param stock if true, it influnces stock market.
     * @returns set of pids of scripts runned.
     */
    async execBatch(hwgw: HackCalculation, stock?: boolean): Promise<number[]> {
        stock ??= false;
        //this.ns.print(`executing batch of ${hwgw.totalThreadCount} threads.`);
        //this.ns.print(`total ram: ${this.totalRam}`);

        this.ns.print(`stage 1: executed ${hwgw.HThreadCount} threads.`);
        const hpid = this.execute("hack",{
            target:this.server.hostname,
            numThread:hwgw.HThreadCount,
            sleepDuration: hwgw.cycleTime - hwgw.hackTime - 1000,
            stock
        });
        await this.scheluder.sleepInWork(250);
        this.ns.print(`stage 2: executed ${hwgw.HWThreadCount} threads.`);
        const wpid = this.execute("weaken",{
            target:this.server.hostname,
            numThread:hwgw.HWThreadCount,
            sleepDuration: hwgw.cycleTime - hwgw.weakenTime - 1000,
            stock
        });
        await this.scheluder.sleepInWork(250);
        this.ns.print(`stage 3: executed ${hwgw.GThreadCount} threads.`);
        const gpid = this.execute("grow",{
            target:this.server.hostname,
            numThread:hwgw.GThreadCount,
            sleepDuration: hwgw.cycleTime - hwgw.growTime - 1000,
            stock
        });
        await this.scheluder.sleepInWork(250);
        this.ns.print(`stage 4: executed ${hwgw.GWThreadCount} threads.`);
        const gwpid = this.execute("weaken",{
            target:this.server.hostname,
            numThread:hwgw.GWThreadCount,
            sleepDuration: hwgw.cycleTime - hwgw.weakenTime - 1000,
            stock
        });
        await this.scheluder.sleepInWork(250);
        return [...hpid, ...wpid, ...gpid, ...gwpid];
    }

    calcHWGW(): HackCalculation {
        const helper = new FormulaHelper(this.ns);
        const cycleTime = 1 + helper.calculateWeakenTime(this.server);
        const ramPerSec = Math.floor(this.totalRam / (cycleTime / 1000));
        const [min,minInfo] = this.getMinThreadCount();
        const minRam = getTotalRamOf(minInfo);
        if(minRam > ramPerSec){
            return minInfo;
        }
        // linear search
        let prev = minInfo;
        for (let i = min+1; i < min + 100; i++) {
            const info = this.calcBatchResource(i);
            if(info === null) break;
            if(getTotalRamOf(info) > ramPerSec){
                return prev;
            }
            prev = info;
        }
        return prev;
    }

    getMinThreadCount():[number,HackCalculation]{
        let minCE = 0;
        let minHC = 0;
        let minInfo: HackCalculation;
        for (let i = 1; i < 100; i++) {
            const info = this.calcBatchResource(i);
            if(info === null) break;
            const ce = info.earnMoneyPerCycle / info.totalThreadCount;
            if (ce >= minCE) {
                minCE = ce;
                minHC = i;
                minInfo = info;
            }
            else break;
        }
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        return [minHC,minInfo!];
    }
}

export function requestHWGW(req: ReserveRequest, target: string): void {
    const key = getKeyOfHWGWBatchManager(target);
    const data = localStorage.getItem(key);
    let requests: ReserveRequest[];
    if (data) {
        requests = JSON.parse(data);
    }
    else {
        requests = [];
    }
    requests.push(req);
    localStorage.setItem(key, JSON.stringify(requests));
}

export function popHWGWRequest(hostname: string): ReserveRequest[]{
    const key = getKeyOfHWGWBatchManager(hostname);
    const text = localStorage.getItem(key)
    if(text === null){
        return [];
    }
    localStorage.setItem(key,"[]");
    return JSON.parse(text) as ReserveRequest[];
}

export async function batchDaemon(ns: NS, hostname: string): Promise<void> {
    ns.disableLog("getBitNodeMultipliers");
    ns.disableLog("sleep");
    const mgr = new HWGWBatchManager(ns, hostname);
    for (;;) {
        const requests = popHWGWRequest(hostname);
        for (const req of requests) {
            mgr.reserve(req.hostname, req.usableRam);
        }
        const hwgw = mgr.calcHWGW();
        if (hwgw !== null && mgr.getUsableTotalRam() >= getTotalRamOf(hwgw)) {
            ns.print(`hwgw: total ${hwgw.totalThreadCount} threads.`);
            mgr.execBatch(hwgw);
        }
        for (let i = 0; i < 4; i++) {
            await mgr.scheluder.tick();
        }
    }
}

function getKeyOfGrowBatchManager(hostname: string): string{
    return `Grow-${hostname}`;
}
export function popGrowRequest(hostname: string): ReserveRequest[]{
    const key = getKeyOfGrowBatchManager(hostname);
    const text = localStorage.getItem(key)
    if(text === null){
        return [];
    }
    localStorage.setItem(key,"[]");
    return JSON.parse(text) as ReserveRequest[];
}
export function requestGrow(req: ReserveRequest, target: string): void {
    const key = getKeyOfGrowBatchManager(target);
    const data = localStorage.getItem(key);
    let requests: ReserveRequest[];
    if (data) {
        requests = JSON.parse(data);
    }
    else {
        requests = [];
    }
    requests.push(req);
    localStorage.setItem(key, JSON.stringify(requests));
}


class GrowManager extends BatchManager {
    constructor(ns: NS, hostname: string) {
        super(ns, hostname);
    }

    async execStraigy(stock: boolean): Promise<void> {
        const helper = new FormulaHelper(this.ns);
        const difficultyMargin = 1;
        if(this.server.hackDifficulty > this.server.minDifficulty + difficultyMargin){
            const weakenThread = helper.calculateWeakenThreadCountToTargetSecurity(
                this.server.hackDifficulty, this.server.minDifficulty);
            const g = this.execute("weaken", {
                target: this.server.hostname,
                numThread: weakenThread,
                sleepDuration: 1,
                stock
            });
            this.ns.print("[PROCESS] weakened");
            const work = async (): Promise<void> => {
                await this.execStraigy(stock);
            }
            this.scheluder.onPidTerminate(g[0], {
                work
            });
        }

        const moneyFactor = 0.8;
        if(this.server.moneyAvailable > this.server.moneyMax * moneyFactor){
            return;
        }
        const totalRam = this.getUsableTotalRam();
        const growThread = Math.floor(totalRam / (getRamOfScript(this.ns, "grow") * 3));
        const weakenThread = growThread * 2;
        const growTime = helper.calculateGrowTime(this.server);
        const weakenTime = helper.calculateWeakenTime(this.server);
        
        this.execute("weaken", {
            target: this.server.hostname,
            numThread: weakenThread,
            sleepDuration: 0,
        });
        await this.scheluder.sleepInWork(250);
        const g = this.execute("grow", {
            target: this.server.hostname,
            numThread: growThread,
            sleepDuration: weakenTime - growTime,
            stock
        });

        const work = async (): Promise<void> => {
            await this.execStraigy(stock);
        }
        this.scheluder.onPidTerminate(g[0], {
            work
        });
    }

    empty(): boolean {
        return this.scheluder.empty();
    }
}

export async function growDaemon(ns: NS, hostname: string): Promise<void> {
    ns.disableLog("getBitNodeMultipliers");
    ns.disableLog("sleep");
    const mgr = new GrowManager(ns, hostname);
    processRequest();
    mgr.execStraigy(false);
    while (!mgr.empty()) {
        processRequest()
        await mgr.scheluder.tick();
    }

    function processRequest(): void{
        const requests = popGrowRequest(hostname);
        for (const req of requests) {
            mgr.reserve(req.hostname, req.usableRam);
        }
    }
}